# utf-8

"""
The initialization for the battlesim package.
"""

""" This block handles the import needs of the package """
hard_dependencies = ("numpy", "matplotlib", "pandas", "scipy", "numba")
missing_dependencies = []

for dependency in hard_dependencies:
    try:
        __import__(dependency)
    except ImportError as e:
        missing_dependencies.append("{0}: {1}".format(dependency, str(e)))

if missing_dependencies:
    raise ImportError("Unable to import required dependencies:\n" + "\n".join(missing_dependencies))
del hard_dependencies, dependency, missing_dependencies
""" Block ends """


# imports
from ._battle import Battle
from ._terrain import Terrain
from ._distributions import Distribution
from . import __legacy

__version__ = "0.3.6"
__name__ = "battlesim"
__doc__ = """
battlesim - Modelling and animating simulated battles between units in Python.
==============================================================================

**battlesim** is a Python package providing TABS (totally-accurate-battle-simulator)-
style combat designed to entertain, inform and acts as a platform for simulation and
modelling within a games-design context. Simulations are designed not only to be
comprehensive and flexible, but also fast by relying on just-in-time compiling.

Main Features
-------------
Here are just a few things that battlesim aims to do well:

    - Formulate your simulation in a few lines of code from scratch.
    - Scales up to thousands (and 10s of thousands) of units
    - Flexibility: unit values are taken from a data file with flexible AI options
    - Performance: Just-in-time compiling can manage thousands of units
    - Visualisation: Animations can be customized to change look-and-feel
"""



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue Oct  1 16:20:15 2019

@author: gparkes

Performs calculations on the 'hit chance' of various attacks.
"""

from numba import njit


@njit
def basic_chance(accuracy, dodge, distances, i, j, global_penalty=15.):
    """
    Hit chance in the range [0..1]
    0 meaning no chance of hitting, 1 meaning perfect accuracy.

    Calculated as:
        chance = <accuracy> * (1 - <enemy dodge>) * (1 - <distance to enemy> / <global penalty>)
    """
    return accuracy[i] * (1. - dodge[j]) * (1. - distances[i] / global_penalty)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Fri Feb 22 14:30:45 2019

@author: gparkes
"""
import numpy as np
import matplotlib.pyplot as plt
from matplotlib import animation
import itertools as it

from battlesim._simulator_fast import frame_columns
from battlesim.utils import check_columns, slice_loop

from matplotlib.lines import Line2D

# all functions to import
__all__ = ["quiver_fight"]


def loop_colors():
    """Provides a looping list of matplotlib colors."""
    return ("red", "blue", "green", "orange", "purple", "brown", "black",
            "cyan", "yellow")


def quiver_fight(frames,
                 terrain=None,
                 allegiance_label={},
                 allegiance_color={},
                 quant_size_map={}):
    """
    Generates an animated quiver plot with units moving around the arena
    and attacking each other. Requires the Frames object as output from a 'battle.simulate()'
    call.

    Units that are alive appear as directional quivers, units that are dead
    appear as crosses 'x'.

    We recommend you use this in conjunction with Jupyter notebook:
        HTML(bsm.quiver_fight(Frames).tojshtml())

    Parameters
    -------
    frames : pd.DataFrame
        The dataframe with each frame step to animate
        Columns included must be: 'x', 'y', 'dir_x', 'dir_y', 'allegiance', 'frame' and 'alive'
    terrain : bsm.Terrain object
        A terrain object to generate and draw from.
    allegiance_label : dict
        maps allegiance in Frames["allegiance"] (k) to a label str (v)
    allegiance_color : dict
        maps allegiance in Frames["allegiance"] (k) to a color str (v)
    quantify_size : dict
        If True, use unit_quant to estimate unit value then set this size to quivers.

    Returns
    ------
    anim : matplotlib.pyplot.animation
        object to animate then from.
    """
    check_columns(frames, frame_columns())

    # set plt.context
    plt.rcParams["animation.html"] = "jshtml"
    # plt.rcParams['animation.writer'] = 'pillow'
    # dataframe
    N_frames = frames.index.unique().shape[0]
    # create plot
    fig = plt.figure(figsize=(8, 6))
    ax = fig.add_subplot(111)
    # plot the terrain underneath
    terrain.plot(ax, alpha=.2)

    # hide axes labels
    ax.get_xaxis().set_visible(False)
    ax.get_yaxis().set_visible(False)

    # use the numerical allegiance.
    allegiances = frames["allegiance"].unique()
    # create defaults if the dictionary size does not match the allegiance flags
    if len(allegiance_label) != allegiances.shape[0]:
        allegiance_label = dict(zip(allegiances.tolist(),
                                    ["team%d" % i for i in it.islice(it.count(1), 0, allegiances.shape[0])]))
    if len(allegiance_color) != allegiances.shape[0]:
        allegiance_color = dict(zip(allegiances.tolist(),
                                    slice_loop(_loop_colors(), allegiances.shape[0])))
    # unique units.
    Uunits = frames["army"].unique()
    if len(quant_size_map) == 0:
        quant_size_map = {k: 1 for k in Uunits}

    combs = list(it.product(allegiances, Uunits))

    """
    Create two groups for each allegiance:
        1. The units that are alive, are arrows.
        2. The units that are dead, are crosses 'x'
    """

    qalive = []
    dead = []

    for a, un in combs:
        f1 = frames.loc[0].query("(allegiance==@a) & (army==@un) & alive")
        team_alive = ax.quiver(f1.x, f1.y, f1.dir_x, f1.dir_y, color=allegiance_color[a], alpha=.5,
                               scale=30 * (quant_size_map[un] + 1.), width=0.015, pivot="mid")
        qalive.append(team_alive)

        team_dead, = ax.plot([], [], 'x', color=allegiance_color[a], alpha=.2, markersize=5.)
        dead.append(team_dead)

    # configure legend, extras.
    # set plot bounds
    xmin, xmax, ymin, ymax = terrain.bounds_
    ax.set_xlim(xmin - .5, xmax + .5)
    ax.set_ylim(ymin - .5, ymax + .5)
    # design custom legend
    custom_lines = [Line2D([0], [0], color=allegiance_color[a], lw=4) for a in allegiances]
    ax.legend(custom_lines, [allegiance_label[a] for a in allegiances], loc="upper right")
    fig.tight_layout()
    plt.close()

    # an initialisation function = to plot at the beginning.
    def _init():
        for j, (_a, _un) in enumerate(combs):
            new_alive = frames.loc[0].query("(allegiance==@_a) & (army==@_un) & alive")
            if len(new_alive) > 0:
                qalive[j].set_UVC(new_alive["dir_x"], new_alive["dir_y"])

        return (*qalive, *dead)

    # animating the graph with step i
    def _animate(i):
        for j, (_a, _un) in enumerate(combs):
            new_alive = frames.loc[i].query("(allegiance == @_a) & (alive) & (army == @_un)")
            new_dead = frames.loc[i].query("(allegiance == @_a) & (not alive) & (army == @_un)")
            if len(new_alive) > 0:
                qalive[j].set_offsets(np.vstack((new_alive["x"], new_alive["y"])).T)
                qalive[j].set_UVC(new_alive["dir_x"], new_alive["dir_y"])
            if len(new_dead) > 0:
                dead[j].set_data(new_dead["x"], new_dead["y"])

        return (*qalive, *dead)

    return animation.FuncAnimation(fig, _animate, init_func=_init,
                                   interval=100, frames=N_frames, blit=True)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Responsible for creating a Battle object.
"""

import numpy as np
import pandas as pd
import warnings
from typing import Dict, Union, Tuple, Callable, List

from . import utils
from . import _target
from .plot import quiver_fight
from . import _ai
from . import _unit_quant
from .__defaults import default_db
from ._simulator_fast import simulate_battle as sim_battle
from ._distributions import Distribution
from ._terrain import Terrain


class Battle(object):
    """
    This 'Battle' object provides the interface for the user of simulating
    a number of Battles.

    Each simulation follows a:
        Load -> Create -> Simulate -> Draw
    flow.
    """

    """###################### INIT FUNCTION #####################################"""

    def __init__(self,
                 db: Union[str, Dict, pd.DataFrame] = default_db(),
                 bounds: Tuple[float, float, float, float] = (0., 10., 0., 10.)):
        """
        Instantiate this object with a filepath leading to

        Parameters
        -------
        db : str, dict or pandas.DataFrame
            If str: Is filepath to the database object
            If dict or pandas.dataFrame: represents actual data.
            Must contain ["Name", "Allegiance", "HP", "Damage", "Accuracy", "Miss", "Movement Speed", "Range"] columns.
            See bsm.defaults.default_db() for example.
        bounds : tuple (4,)
            The left, right, top and bottom bounds of the battle. Units cannot
            leave these bounds.
        """
        # assign with checks
        self.db_ = db
        self._D = []
        self._M = None
        self._sim = None
        self.db_.index = self.db_.index.str.lower()
        # initialise a terrain
        self._T = Terrain(bounds, res=.1, form=None)
        # create internal dictionary called 'army group' or ag
        self._ag = {}

    """####################### HIDDEN FUNCTIONS ##############################"""

    def _dataset(self, n: int):
        return np.zeros(n, dtype=[
            ("team", np.uint8, 1), ("utype", np.uint8, 1), ("pos", np.float32, 2), ("hp", np.float32, 1),
            ("armor", np.float32, 1), ("range", np.float32, 1), ("speed", np.float32, 1), ("acc", np.float32, 1),
            ("dodge", np.float32, 1), ("dmg", np.float32, 1), ("target", np.int32, 1),
            ("group", np.uint8, 1)
        ])

    def _is_instantiated(self):
        if self.M_ is None:
            raise AttributeError("'create_army' has not been called - there are no units.")

    def _is_simulated(self):
        if self.sim_ is None:
            raise AttributeError("No simulation has occurred, no presense of battle.sim_ object.")

    def _determine_target_ai(self, t, pkg=_target):
        self._is_instantiated()
        if isinstance(t, str):
            if t in pkg.get_function_names():
                return [t] * self.n_armies_
            else:
                raise ValueError("value '{}' not found in {}".format(t, pkg.get_function_names()))
        elif isinstance(t, (list, tuple)) and len(t) == self.n_armies_:
            utils.check_in_list(pkg.get_function_names(), t)
            utils.check_list_type(t, str)
            return t
        else:
            raise AttributeError("target ai type must be a type of [str, list, tuple].")

    def _determine_ai_mapping(self, t, pkg=_target):
        mm = pkg.get_map_functions()
        r_armies = range(self.n_armies_)
        army_to_fname = dict(zip(r_armies, t))
        return dict(zip(r_armies, [mm[army_to_fname[f]] for f in army_to_fname]))

    def _plot_simulation(self, func: Callable):
        labels = self.allegiances_.to_dict()
        cols = utils.slice_loop(_simplot.loop_colors(), len(self.allegiances_))
        # quantify size by value
        qscore = _unit_quant.rank_score(self.db_).reset_index(drop=True).to_dict()

        # call plotting function - with or without terrain
        if self.T_ is not None:
            Q = func(self.sim_, self.T_, labels, cols, qscore)
        else:
            Q = func(self.sim_, None, labels, cols, qscore)
        return Q

    def _initialize_position(self):
        """ Given self.D_, instantiation, we should be able to map positions """
        self._is_instantiated()
        for i, (u, n, (start, end), dist) in enumerate(zip(self._unit_roster, self._unit_n, self._segments, self.D_)):
            self.M_["pos"][start:end] = dist.sample(n)

    def _get_bounds_from_M(self):
        xmin, xmax = self.M_["pos"][:, 0].min(), self.M_["pos"][:, 0].max()
        ymin, ymax = self.M_["pos"][:, 1].min(), self.M_["pos"][:, 1].max()
        return np.floor(xmin), np.ceil(xmax), np.floor(ymin), np.ceil(ymax)

    def _check_bounds_to_M(self, bounds: Tuple[float, float, float, float]):
        xmin, xmax, ymin, ymax = self._get_bounds_from_M()
        if bounds[0] > xmin:
            raise ValueError("xmin bounds value: {} > unit bound {}".format(bounds[0], xmin))
        if bounds[1] < xmax:
            raise ValueError("xmax bounds value: {} < unit bound {}".format(bounds[1], xmax))
        if bounds[2] > ymin:
            raise ValueError("ymin bounds value: {} > unit bound {}".format(bounds[2], ymin))
        if bounds[3] < ymax:
            raise ValueError("ymax bounds value: {} < unit bound {}".format(bounds[3], ymax))

    def _sample_dist(self, D, segments):
        for (u, n), (start, end) in zip(self.army_set_, segments):
            self.D_.append(D)
            self.M_["pos"][start:end] = D.sample(n)

    def _assign_initial_targets(self):

        f_dict = _target.get_global_map_functions()

        for group, (start, end), func, team in zip(range(self.n_armies_), self._segments, self.init_ai_, self._teams):
            mod_func = "global_" + func
            self.M_["target"][start:end] = f_dict[mod_func](
                self.M_["pos"], self.M_["hp"], self.M_["team"], self.M_["group"], group
            )

    """---------------------- PUBLIC ATTRIBUTES AND ATTR METHODS ----------------------------------------"""

    @property
    def composition_(self) -> pd.DataFrame:
        """Determines the composition of the Battle."""
        self._is_instantiated()
        d = {"unit": [name for name, _ in self.army_set_],
             "allegiance": [self.db_.loc[u, "Allegiance"] for u, _ in self.army_set_],
             "n": [n for _, n in self.army_set_],
             "position": ["{} ({:0.1f}, {:0.1f})".format(dist.dist_, dist.mean_[0], dist.mean_[1]) for dist in self.D_],
             "init_ai": self.init_ai_,
             "rolling_ai": self.rolling_ai_,
             "decision_ai": self.decision_ai_}
        return pd.DataFrame(d)

    @property
    def n_allegiance_(self):
        """Determines the number of teams present in the fight."""
        self._is_instantiated()
        d = {"allegiance": [self.db_.loc[u, "Allegiance"] for u, _ in self.army_set_],
             "n": [n for _, n in self.army_set_]}
        return pd.DataFrame(d).groupby("allegiance")["n"].sum()

    @property
    def bounds_(self) -> Tuple[float, float, float, float]:
        """Determine the bounds of the fight using the Terrain."""
        return self.T_.bounds_

    @bounds_.setter
    def bounds_(self, b: Tuple[float, float, float, float]):
        if self.M_ is None:
            warnings.warn("bounds {} set before units are initialised - some may be out-of-bounds".format(b),
                          UserWarning)
            self.T_.bounds_ = b
        else:
            self._check_bounds_to_M(b)
            self.T_.bounds_ = b

    @property
    def M_(self) -> np.ndarray:
        """The raw data set underlying."""
        return self._M

    @property
    def sim_(self):
        """The simulation object."""
        return self._sim

    @property
    def db_(self) -> pd.DataFrame:
        """The datafiles storing information on each unit."""
        return self._db

    @db_.setter
    def db_(self, db_n):
        if isinstance(db_n, str):
            self._db = utils.import_and_check_unit_file(db_n)
        elif isinstance(db_n, dict):
            self._db = pd.DataFrame(db_n)
            utils.check_unit_file(self._db)
            utils.preprocess_unit_file(self._db)
        elif isinstance(db_n, pd.DataFrame):
            self._db = db_n.copy()
            utils.check_unit_file(self._db)
            utils.preprocess_unit_file(self._db)
        else:
            raise ValueError("'db' must be of type [str, dict, pd.DataFrame], not {}".format(type(db_n)))

    @property
    def T_(self) -> Terrain:
        """Attribute to the Terrain object."""
        return self._T

    @property
    def army_set_(self):
        """A list of unit rosters and counts."""
        self._is_instantiated()
        return tuple(zip(self._unit_roster, self._unit_n))

    @property
    def n_armies_(self) -> int:
        """The number of army types."""
        self._is_instantiated()
        return len(self._unit_roster)

    @property
    def N_(self) -> int:
        """The total number of units."""
        self._is_instantiated()
        return sum(self._unit_n)

    @property
    def D_(self):
        """A set of Distribution objects."""
        return self._D

    @D_.setter
    def D_(self, d):
        """ where d is a list/tuple of distribution, str, or dict """
        self._is_instantiated()
        self._D = []
        if isinstance(d, (str, Distribution)):
            self._D = [Distribution(d) for _ in range(self.n_armies_)]
        elif isinstance(d, dict):
            self._D = [Distribution(**d) for _ in range(self.n_armies_)]
        elif isinstance(d, (list, tuple)):
            for dist in d:
                if isinstance(dist, (str, Distribution)):
                    self._D.append(Distribution(dist))
                elif isinstance(dist, dict):
                    self._D.append(Distribution(**dist))
        else:
            raise TypeError("'d' must be of type [list, tuple, str, Distribution, dict]")

    @property
    def allegiances_(self):
        """The list of allegiances."""
        return self.db_[["Allegiance", "allegiance_int"]].set_index("allegiance_int").drop_duplicates().squeeze()

    @property
    def init_ai_(self):
        """The initialization AI chosen for selecting targets."""
        return self._init_ai

    @init_ai_.setter
    def init_ai_(self, ia):
        self._init_ai = self._determine_target_ai(ia)

    @property
    def rolling_ai_(self):
        """The AI chosen whenever an enemy unit dies."""
        return self._rolling_ai

    @rolling_ai_.setter
    def rolling_ai_(self, ra):
        self._rolling_ai = self._determine_target_ai(ra)

    @property
    def decision_ai_(self):
        """The AI chosen to govern the actions of each army set."""
        return self._decision_ai

    @decision_ai_.setter
    def decision_ai_(self, da):
        self._decision_ai = self._determine_target_ai(da, pkg=_ai)

    ################### HIDDEN ATTRIBUTE #################################

    @property
    def _segments(self) -> Tuple[str, int]:
        return utils.get_segments(self.army_set_)

    @property
    def _teams(self) -> np.ndarray:
        self._is_instantiated()
        return np.asarray([self.db_.loc[u, "allegiance_int"] for u in self._unit_roster])

    """--------------------------------- PUBLIC FUNCTIONS ------------------------------------------------"""

    def create_army(self, army_set: Tuple[str, int]):
        """
        Armies are groupings of (<'Unit Type'>, <number of units>). You can
        create one or more of these.

        We make use of the dataset (`db`) with army_set.

        We create the 'M' matrix, which is directly fed into any 'simulation' function.

        Parameters
        -------
        army_set : list of 2-tuple
            A list of 'army groups' given as ('Unit Type', number of units)

        Returns self
        -------
        self
        """
        # assign
        utils.is_twotuple(army_set, str, (int, np.int, np.int64))
        # assign unit roster, n for roster
        self._unit_roster = [u.lower() for u, _ in army_set]
        self._unit_n = [n for _, n in army_set]

        self._M = self._dataset(sum(self._unit_n))
        # check that groups exist in army_set
        utils.check_groups_in_db(self.army_set_, self.db_)

        # capture distributions
        self._D = [Distribution("normal", loc=i * 2., scale=1.) for i in range(self.n_armies_)]

        # set initial values.
        for i, (u, n, (start, end)) in enumerate(zip(self._unit_roster, self._unit_n, self._segments)):
            self.M_["utype"][start:end] = np.argwhere(self.db_.index == u).flatten()[0]
            self.M_["team"][start:end] = self.db_.loc[u, "allegiance_int"]
            self.M_["group"][start:end] = i
            self.M_["hp"][start:end] = self.db_.loc[u, "HP"]
            self.M_["armor"][start:end] = self.db_.loc[u, "Armor"]
            self.M_["range"][start:end] = self.db_.loc[u, "Range"]
            self.M_["speed"][start:end] = self.db_.loc[u, "Movement Speed"]
            self.M_["dodge"][start:end] = self.db_.loc[u, "Miss"] / 100.
            self.M_["acc"][start:end] = self.db_.loc[u, "Accuracy"] / 100.
            self.M_["dmg"][start:end] = self.db_.loc[u, "Damage"]

        self._initialize_position()

        # modify bounds
        self.bounds_ = self._get_bounds_from_M()

        # initialise AIs as nearest.
        self.set_initial_ai("nearest")
        self.set_rolling_ai("nearest")
        # main AI options
        self.set_decision_ai("aggressive")

        return self

    def apply_position(self, distributions):
        """
        Assign locations to each 'army set' using a distribution from the battlesim.Distribution object.

        e.g

        battle.apply_distribution("gaussian")
        battle.apply_distribution(bsm.Distribution("normal",loc=0,scale=1))
        battle.apply_distribution([bsm.Distribution("beta"), bsm.Distribution("normal")])
        battle.apply_distribution([
            {"name":"beta", "x_loc":0., "y_loc":1.},
            {"dist":"normal", "x_loc":10., "y_loc": 10.}
        ])

        Parameters
        -------
        distributions : str, battlesim.Distribution or list/tuple of battlesim.Distribution/dict.
            The distribution(s) corresponding to each group.
            e.g
            str : distribution name for all army sets
            dict : distribution name and parameters for all army sets
            Distribution : object for all army sets
            list/tuple:
                Distribution : object for each army set
                dict : name and parameters for each army set
                str : dist name for each army set

        Returns
        -------
        self
        """
        # assign and check
        self.D_ = distributions
        # assign positions
        self._initialize_position()
        # modify bounds
        self.bounds_ = self._get_bounds_from_M()
        return self

    def apply_terrain(self, t=None, res: float = .1):
        """
        Applies a Z-plane to the map that the Battle is occuring on by creating
        a bsm.Terrain object.

        Parameters
        -------
        t : str or bsm.Terrain
            Choose from [None, 'grid', 'contour']. Default is None. Contour looks
            the best. Decides how big/resolution to make the terrain based on the
            initialized positions of units.
        res : float
            The resolution to use for the map

        Returns
        -------
        self
        """
        self._is_instantiated()

        if t in [None, "grid", "contour"]:
            # add function to t
            self.T_.res_ = res
            self.T_.form_ = t
            return self
        elif isinstance(t, Terrain):
            self._T = t
            return self
        else:
            raise ValueError("'t' must be [grid, contour, None]")

    def set_initial_ai(self, func_names: Union[str, List[str]]):
        """
        Set the AI decisions for choosing a target initially. Pass a function name from bsm.ai
        for each 'army set'.

        Parameters
        --------
        func_names : str or list of str
            The AI function to choose for initial targets for each army set. Can choose from:
                ['random', 'pack', 'nearest']

        Returns
        -------
        self
        """
        self.init_ai_ = func_names
        return self

    def set_rolling_ai(self, func_names: Union[str, List[str]]):
        """
        Set the AI decisions for choosing a target rolling through the simulation.
        Pass a function name from bsm.ai for each 'army set'.

        Parameters
        --------
        func_names : str or list of str
            The AI function to choose for targets when their target dies
            for each army set. Can choose from: ['random', 'pack', 'nearest']

        Returns
        -------
        self
        """
        self.rolling_ai_ = func_names
        # map these strings to actual functions, ready for simulate.
        self._rolling_map = self._determine_ai_mapping(self.rolling_ai_)
        return self

    def set_decision_ai(self, decision: Union[str, List[str]]):
        """
        Sets the over-arching AI choices for each 'army group'. By default they
        choose an 'aggressive' stance.

        WARNING: This function CAN override the choices made for the 'initial'
        and 'rolling' AI decisions, based on it's functionality. This function
        is therefore superior to those options.

        Parameters
        --------
        decision : str or list of str
            If str, applies that option to all army groups. Choose from
            [aggressive, hit_and_run]

        Returns
        -------
        self
        """
        self.decision_ai_ = decision
        # map these strings to actual functions, ready for simulate.
        self._decision_map = self._determine_ai_mapping(self.decision_ai_, pkg=_ai)
        return self

    def set_bounds(self, bounds: Tuple[float, float, float, float]):
        """
        Sets the boundaries of the Battle. If not initialised, this is OK but may
        produce errors down-the-line.

        Parameters
        -------
        bounds : tuple, list (4,)
            the (left, right, top, bottom) dimensions of the simulation.

        Returns
        -------
        self
        """
        self.bounds_ = bounds
        return self

    """ ----------------------------- SIMULATION ----------------------------- """

    def simulate(self, **kwargs):
        """
        Runs the 'simulate_battle' algorithm. Creates and passes a copy to simulate..

        Returns pd.DataFrame of frames.
        """
        self._is_instantiated()
        # check for multiple teams
        if np.unique(self._teams).shape[0] <= 1:
            warnings.warn("Simulation halted - There is only one team present.", UserWarning)
            return self.sim_
        # set the flat terrain if it doesn't exist
        self.T_.generate()
        # firstly assign initial AI targets.
        self._assign_initial_targets()
        # we cache a copy of the sim as well for convenience
        self._sim = sim_battle(np.copy(self.M_),
                               self.T_,
                               self._rolling_map,
                               self._decision_map,
                               ret_frames=True,
                               **kwargs)
        return self.sim_

    def simulate_k(self, k: int = 10, **kwargs):
        """
        Runs the 'simulate_battle' algorithm 'k' times. Creates and passes a copy
        to simulate.

        Parameters
        --------
        k : int
            The number of iterations. Must be at least 1.
        **kwargs : dict
            keyword arguments to pass to simulate_battle.

        Returns
        -------
        runs : pd.DataFrame
            The iteration (k), with the team victorious, and number of units remaining

        Returns the victory for each k iteration, for each team.
        """
        self._is_instantiated()

        if k < 1:
            raise ValueError("'k' must be at least 1")
        else:
            # check for multiple teams
            if np.unique(self._teams).shape[0] <= 1:
                warnings.warn("Simulation halted - There is only one team present.", UserWarning)
                return self.sim_

            runs = np.zeros((k, 2), dtype=np.int64)
            # generate new terrain
            self.T_.generate()
            for i in range(k):
                # firstly assign initial AI targets.
                self._assign_initial_targets()
                # run simulation
                team_counts = sim_battle(np.copy(self.M_),
                                         self.T_,
                                         self._rolling_map,
                                         self._decision_map,
                                         ret_frames=False,
                                         **kwargs)
                runs[i, :] = team_counts
            return pd.DataFrame(runs, columns=self.allegiances_.values)

    """ ------------ CONVENIENCE PLOTTING FUNCTIONS ---------------------- """

    def sim_jupyter(self,
                    func: Callable = quiver_fight,
                    create_html: bool = False):
        """
        This convenience method uses any saved 'sim_' object to generate the code
        to output to a Jupyter Notebook. Once must simply then do:

            HTML(battle.sim_jupyter())

        And hey presto, it should all work!

        Parameters
        --------
        func : function, optional
            The plot function to call, by default is bsm.quiver_fight()
        create_html : bool, optional
            Decides whether to return the object directly, or create HTML to then use HTML()

        Returns
        -------
        s : str/object
            HTML code to feed into HTML(s)
        """
        self._is_simulated()
        # call plotting function - with
        Q = self._plot_simulation(func)

        if create_html:
            return Q.to_jshtml()
        else:
            return Q

    def sim_export(self,
                   filename: str = "example_sim.gif",
                   func: Callable = quiver_fight,
                   writer: str = "pillow"):
        """
        This convenience method uses any saved 'sim_' object to generate the code
        to export into a gif file.

        Parameters
        -------
        filename : str, optional
            The name of the file to output. Must end in .gif
        func : function, optional
            The plot function to call, by default is bsm.quiver_fight()
        writer : str, optional
            The type of writer to pass to funcanimation.save(). This might
            need to be tweaked on your system.
            Accepts ['imagemagick', 'ffmpeg', 'pillow']

        Returns
        -------
        None
        """
        self._is_simulated()
        # append to end if not present
        if not filename.endswith(".gif"):
            filename.append(".gif")

        # call simulation
        Q = self._plot_simulation(func)

        # save
        Q.save(filename, writer=writer)
        return

    """ ---------------------- MISC --------------------------------------- """

    def __repr__(self) -> str:
        if self.M_ is None:
            return "bsm.Battle(init=False)"
        elif self.sim_ is None:
            return "bsm.Battle(init=True, n_armies={}, simulated=False)".format(self.n_armies_)
        else:
            return "bsm.Battle(init=True, n_armies={}, simulated=True)".format(self.n_armies_)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue Feb 19 15:48:25 2019

@author: gparkes

This class handles the primary simulator functions given some data.
"""
import pandas as pd
import numpy as np

from . import _jitcode

############################################################################

__all__ = ["simulate_battle"]


def frame_columns():
    """Column names w.r.t frame dataframe."""
    return "army", "allegiance", "alive", "armor", "hp", "x", "y", "dir_x", "dir_y"


def _copy_frame(Frames, M, i):
    # copy over data from M into frames.
    Frames["frame"][i] = i
    Frames["pos"][i] = M["pos"]
    Frames["target"][i] = M["target"]
    Frames["hp"][i] = M["hp"]
    Frames["armor"][i] = M["armor"]
    # create direction norm
    dnorm = _jitcode.direction_norm(M["pos"][M["target"]] - M["pos"])
    Frames["dpos"][i] = dnorm
    Frames["team"][i] = M["team"]
    Frames["utype"][i] = M["utype"]
    return


def _convert_to_pandas(frames):
    """
    Given the 'frames' from fast_simulate_battle, return a pd.Dataframe
    """
    steps, units = frames.shape
    # use frame as index to reduce memory.
    DF = pd.concat([pd.DataFrame({
        "army": frames["utype"][s],
        "allegiance": frames["team"][s],
        "alive": frames["hp"][s] > 0,
        "hp": np.clip(frames["hp"][s], a_min=0., a_max=None),
        "armor": np.clip(frames["armor"][s], a_min=0, a_max=None),
        "x": frames["pos"][s][:, 0],
        "y": frames["pos"][s][:, 1],
        "dir_x": frames["dpos"][s][:, 0],
        "dir_y": frames["dpos"][s][:, 1]
    }, index=frames["frame"][s]) for s in range(steps)])
    DF.index.name = "frame"
    return DF


def simulate_battle(M,
                    terrain,
                    target_map,
                    decision_map,
                    max_step=100,
                    ret_frames=True):
    """
    Given a Numpy Matrix of units, simulate a fight.

    This uses a matrix M which is **heterogenous**; that is to say that
    it has named columns [pos, target, hp, dpos, team, group] which
    helps interpretability at the cost of some processing time.

    Parameters
    --------
    M : np.ndarray (units, )
        A heterogenous matrix containing data values for units
    terrain : bsm.Terrain object
        Terrain object containing the bounds.
    target_map : dict
        A dictionary mapping groups (k) to a bsm.target.* function (v)
    decision_map : dict
        A dictionary mapping groups (k) to a bsm.ai.* function (v)
    max_step : int
        The maximum number of steps
    ret_frames : bool
        If True, save each frame, else, return the allegiance that is victorious.

    Returns
    -------
    frames : np.ndarray (frame, :)
        Each frame is a numpy.ndmatrix.
    """
    t = 0
    running = True
    teams = np.unique(M["team"])
    # unpack bounds
    xmin, xmax, ymin, ymax = terrain.bounds_

    # flatten X, Y
    X_m, Y_m = terrain.get_flat_grid()
    # repeat X_m, Y_m for n units.
    X_m = np.repeat(X_m, M.shape[0]).reshape(X_m.shape[0], M.shape[0])
    Y_m = np.repeat(Y_m, M.shape[0]).reshape(Y_m.shape[0], M.shape[0])
    # height
    Z_m = terrain.Z_

    if ret_frames:
        frames = np.zeros(
            (max_step + 1, M.shape[0]),
            dtype=[("frame", np.int64, 1), ("pos", np.float32, 2), ("target", np.int32, 1),
                   ("hp", np.float32, 1), ("armor", np.float32, 1), ("dpos", np.float32, 2), ("team", np.uint8, 1),
                   ("utype", np.uint8, 1)
            ]
        )

    while (t < max_step) and running:

        # copy a frame
        if ret_frames:
            _copy_frame(frames, M, t)

        # perform a boundary check.
        _jitcode.boundary_check(xmin, xmax, ymin, ymax, M["pos"])
        # list of indices per unit for which tile they are sitting on (X, Y)
        X_t_ind = np.argmin(np.abs(M["pos"][:, 0] - X_m), axis=0)
        Y_t_ind = np.argmin(np.abs(M["pos"][:, 1] - Y_m), axis=0)

        """# pre-compute the direction derivatives and magnitude/distance for each unit to it's target in batch."""
        dir_vec = M["pos"][M["target"]] - M["pos"]
        dists = _jitcode.euclidean_distance(dir_vec)
        """precompute enemy and ally target listings"""
        enemy_targets = [np.argwhere((M["hp"] > 0) & (M["team"] != T)).flatten() for T in teams]
        ally_targets = [np.argwhere((M["hp"] > 0) & (M["team"] == T)).flatten() for T in teams]
        """# pre-compute the 'luck' of each unit with random numbers."""
        round_luck = np.random.rand(M.shape[0], 2)

        # iterate over units and check their life, target.
        for i in range(M.shape[0]):
            if M["hp"][i] > 0.:
                dm = decision_map[M["group"][i]]
                # AI-based decision for attack/defend.
                running = (dm(
                    # variables
                    M["pos"], M["speed"], M["range"], M["acc"], M["dodge"],
                    M["target"], M["dmg"], M["hp"], M["armor"], round_luck, dists, dir_vec,
                    M["team"], target_map[M["group"][i]],
                    enemy_targets[M["team"][i]], ally_targets[M["team"][i]],
                    Z_m, X_t_ind, Y_t_ind, i
                ))

        t += 1

    if ret_frames:
        _copy_frame(frames, M, t)
        return _convert_to_pandas(frames[:t])
    else:
        return np.asarray([np.argwhere((M["hp"] > 0) & (M["team"] == T)).flatten().shape[0] for T in teams])



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue Oct  8 11:12:15 2019

@author: gparkes
"""

from numba import njit

"""
Includes a bonus damage when attacking downhill.
*New in 0.3.5*: Factoring in Armor
For simplicity, we treat Armor as an extra health-pool for now.

Total Damage (without armor) = Base Damage * ((terrain[i] - terrain[j]) / 2) + 1
"""


@njit
def bonus_terrain_damage(Z, Z_xi, Z_yi, i, j):
    """Calculates the bonus terrain damage done."""
    return ((Z[Z_xi[i], Z_yi[i]] - Z[Z_xi[j], Z_yi[j]]) / 2.) + 1.


@njit
def basic(hp, armor, target, damage, Z, Z_xi, Z_yi, i):
    """Determines basic damage output to target i."""
    # cache base damage.
    dmg = (damage[i] * bonus_terrain_damage(Z, Z_xi, Z_yi, i, target[i]))
    if armor[target[i]] > 0:
        # deal damage to armor (and a little bit to HP)
        diff = armor[target[i]] - dmg
        if diff < 0:
            hp[target[i]] += diff
        armor[target[i]] -= dmg
    else:
        hp[target[i]] -= dmg



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue Oct  8 11:07:18 2019

@author: gparkes
"""
from numba import njit

from . import _move
from . import _hit
from . import _damage


def get_function_names():
    """Returns the function names."""
    return "aggressive", "hit_and_run"


def get_functions():
    """Returns the functions themselves."""
    return aggressive, hit_and_run


def get_map_functions():
    """Maps the function names to the functions."""
    return dict(zip(get_function_names(), get_functions()))


__all__ = get_function_names()

"""
Parameters:
    pos: (np.ndarray, (n, 2), float all) positions of every unit
    speed: (np.ndarray, (n,) float all) base speed of every unit
    mrange: (np.ndarray, (n,) float all) base range of every unit
    acc: (np.ndarray, (n,) float all) accuracy of every unit
    dodge: (np.ndarray, (n,) float all) dodge of every unit
    target: (np.ndarray, (n,) int all) id of every target
    dmg: (np.ndarray, (n,) float all) base damage of every unit
    hp: (np.ndarray, (n,) float all) current HP of every unit
    armor: (np.ndarray, (n,) float all) current armor of enemy unit
    luck: (np.ndarray, (n,2) float all) luck roll of every unit
    distances: (np.ndarray, (n,) float all) unit distance to target
    dd: (np.ndarray, (n,) float all) directional derivative to target
    team: (np.ndarray, (n, ) int all) team ID of every unit
    target_f: (function) target function.
    enemies: (np.ndarray, (e,) int all) IDs of every enemy unit
    allies: (np.ndarray, (a,) int all) IDs of every allied unit
    Z: (np.ndarray, (res, res) float all) the terrain of the map
    Z_xi: (np.ndarray, (n,) int all) the x-index of Z-terrain tile
    Z_yi: (np.ndarray, (n,) int all) the y-index of Z-terrain tile
    i : (int) the current unit's index
"""


@njit
def _height_modification(h):
    """
    height modifier follows:
        h_{mod} = 1 + (h^2 / 3), h in [0, 1]
    """
    return (h ** 2 / 3.) + 1.


@njit
def _select_enemy(f, pos, targets, hp, enemies, allies, i):
    if hp[targets[i]] <= 0:
        if enemies.shape[0] > 0:
            t = f(pos, hp, enemies, allies, i)
            if t != -1:
                targets[i] = t
                return True
            else:
                return False
        else:
            return False
    else:
        return True


@njit
def _rng_terr(mrange, Z, Z_xi, Z_yi, i):
    """ The range of the unit, modified by the terrain measurement. Returns float. """
    return mrange[i] * _height_modification(Z[Z_xi[i], Z_yi[i]])


""" -------------------------- ACCESSED FUNCTIONS -----------------------------------------"""


@njit
def aggressive(pos, speed, mrange, acc, dodge, targets,
               dmg, hp, armor, luck, distances, dd, team, target_f,
               enemies, allies, Z, Z_xi, Z_yi, i):
    """
    This basic AI looks whether the current unit i is in range of it's target'
    and if it isn't, moves towards it until it is, then attacks.
    """

    """# use ai_map to dictionary-map the group number to the appropriate AI function"""
    """ Arguments: positions, targets, hp, enemies, allies, index, [extras]"""
    if _select_enemy(target_f, pos, targets, hp, enemies, allies, i):
        # if not in range, move towards target, or hit a chance (5%) and move forward anyway.
        if distances[i] > _rng_terr(mrange, Z, Z_xi, Z_yi, i) or luck[i, 1] < 0.05:
            """# move unit towards attacking enemy."""
            pos[i, :] += _move.to_enemy(speed, dd, distances, Z, Z_xi, Z_yi, i)
            return True
        else:
            """# calculate the chance of hitting the opponent"""
            h_chance = _hit.basic_chance(acc, dodge, distances, i, targets[i])
            """if hit chance overcomes round luck.. deal damage to HP."""
            if h_chance > luck[i, 0]:
                _damage.basic(hp, armor, targets, dmg, Z, Z_xi, Z_yi, i)
            return True
    else:
        return False


@njit
def hit_and_run(pos, speed, mrange, acc, dodge, targets,
                dmg, hp, armor, luck, distances, dd, team, target_f,
                enemies, allies, Z, Z_xi, Z_yi, i):
    """
    This AI option sees if its range/movement is greater than it's enemy, and if it is, it
    performs hit-and-run on it's opponent.
    """
    # assign target enemy
    if _select_enemy(target_f, pos, targets, hp, enemies, allies, i):

        if (speed[i] > speed[targets[i]]) and \
                (_rng_terr(mrange, Z, Z_xi, Z_yi, i) > _rng_terr(mrange, Z, Z_xi, Z_yi, targets[i])):

            # if we're out of range, move towards
            if distances[i] > _rng_terr(mrange, Z, Z_xi, Z_yi, i):
                """# move towards unit."""
                pos[i, :] += _move.to_enemy(speed, dd, distances, Z, Z_xi, Z_yi, i)
                return True
            # else if the enemy is in range, back off
            elif distances[i] < _rng_terr(mrange, Z, Z_xi, Z_yi, targets[i]):
                """# move directly away from unit."""
                pos[i, :] -= _move.to_enemy(speed, dd, distances, Z, Z_xi, Z_yi, i)
                return True
            else:
                """# so we're in range, the enemy is not, attack."""
                h_chance = _hit.basic_chance(acc, dodge, distances, i, targets[i])
                if h_chance > luck[i, 0]:
                    _damage.basic(hp, armor, targets, dmg, Z, Z_xi, Z_yi, i)
                return True
        else:
            # otherwise just perform an 'aggressive' model.
            return aggressive(pos, speed, mrange, acc, dodge, targets,
                              dmg, hp, armor, luck, distances, dd,
                              team, target_f, enemies, allies,
                              Z, Z_xi, Z_yi, i)
    else:
        return False


def defensive(pos, speed, mrange, acc, dodge, targets,
              dmg, hp, armor, luck, distances, dd, team, target_f,
              enemies, allies, Z, Z_xi, Z_yi, i):
    """
    This AI option attempts to find a nearby high hill and sit on it, waiting for
    a nearby enemy.
    """
    # use target index to see if we can reach the hill in-time.
    return NotImplemented



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Fri Oct  4 15:34:25 2019

@author: gparkes
"""

import matplotlib.pyplot as plt
from matplotlib import animation
import numpy as np
import pandas as pd
import itertools as it
import random
import time
from copy import deepcopy
import os

__units__ = ["Unit1"]
__simulations__ = ["basic_simulation%d" % (i + 1) for i in range(4)]
__animate__ = ["basic_animate%d" % (i + 1) for i in range(3)]
__math__ = ["euclidean", "magnitude", "deriv_norm", "dudt"]
__ai__ = ["find_random_enemy", "find_nearest_enemy"]
__plot__ = ["plot_quiver"]

# __all__ = __units__ + __simulations__ + __math__ + __ai__ + __plot__
__all__ = []

"""############################ UNIT CLASSES ############################################"""


class Unit1(object):
    """This first class defines a Basic Python object unit."""

    def __init__(self, db, name):
        """
        Defines a unit, based on examples in the 'db' file, given a name. Assumes the name
        is the index in db.
        """
        # check that name is in db.index
        if name not in db.index:
            raise ValueError("unit {} must be in {}".format(name, db.index))
        self.name = name
        self.hp = db.loc[name, "HP"]
        self.dmg = db.loc[name, "Damage"]
        self.range = db.loc[name, "Range"]
        self.speed = db.loc[name, "Movement Speed"]
        self.team = db.loc[name, "allegiance_int"]
        self.team_name = db.loc[name, "Allegiance"]

        # define position
        self.pos = np.array([0., 0.])
        # directional derivative from target.
        self.dd = np.array([0., 0.])
        # distance from target
        self.dist = 0.
        # define target to aim for
        self.target = None

    # a basic method to assign position
    def set_position(self, pos_params):
        """Sets the position using pos_parameters of mean and variance."""
        self.pos = np.random.normal(*pos_params, size=(2,))
        return self


"""########################## SIMULATIONS ################################################"""


def basic_simulation1(units, max_step=100):
    """The first basic simulation."""
    t = 0
    running = True

    # while loop
    while running and (t < max_step):
        # iterate over units.
        for i, u in enumerate(units):
            # if we're alive...
            if u.hp > 0:
                # euclidean distance
                u.dd = deriv_norm(u.pos, u.target.pos)
                u.dist = euclidean(u.pos, u.target.pos)
                # now, if the unit is not in range, then move towards it
                if u.dist > u.range:
                    # speed modifier added, others ignored from above.
                    u.pos += dudt(u.pos, u.target.pos, u.speed)
                else:
                    # perform damage to enemy health.
                    u.target.hp -= u.dmg
        t += 1


def basic_simulation2(units, max_step=50):
    """The second basic simulation."""
    t = 0
    running = True

    steps = []

    # while loop
    while running and (t < max_step):
        # iterate over units.
        for i, u in enumerate(units):
            # if we're alive...
            if u.hp > 0:
                # euclidean distance
                u.dd = deriv_norm(u.pos, u.target.pos)
                u.dist = euclidean(u.pos, u.target.pos)
                # now, if the unit is not in range, then move towards it
                if u.dist > u.range:
                    # speed modifier added, others ignored from above.
                    u.pos += dudt(u.pos, u.target.pos, u.speed)
                else:
                    # perform damage to enemy health.
                    u.target.hp -= u.dmg

        # save
        steps.append(deepcopy(units))
        t += 1

    return steps


def basic_simulation3(units, max_step=50):
    """The third basic simulation."""
    # create a copy to prevent overriding
    units = deepcopy(units)
    t = 0
    running = True

    steps = []

    # while loop
    while running and (t < max_step):
        # iterate over units.
        for i, u in enumerate(units):
            # if we're alive...
            if u.hp > 0:

                # wait a second. what if our enemy is dead..?
                if u.target.hp <= 0.:
                    # find a new random enemy.
                    target = find_random_enemy(u, units)
                    if target != -1:
                        u.target = target
                    else:
                        # no more enemies - stop!
                        running = False

                """Only do all this once we have a valid target..."""
                # euclidean distance
                u.dd = deriv_norm(u.pos, u.target.pos)
                u.dist = euclidean(u.pos, u.target.pos)
                # now, if the unit is not in range, then move towards it
                if u.dist > u.range:
                    # speed modifier added, others ignored from above.
                    u.pos += dudt(u.pos, u.target.pos, u.speed)
                else:
                    # perform damage to enemy health.
                    u.target.hp -= u.dmg

        # save
        steps.append(deepcopy(units))
        # add step
        t += 1

    return steps


def basic_simulation4(units, max_step=50):
    """The fourth basic simulation."""
    # create a copy of the units
    units = deepcopy(units)
    # begin t
    t = 0
    running = True

    steps = []

    # while loop
    while running and (t < max_step):
        # iterate over units.
        for i, u in enumerate(units):
            # if we're alive...
            if u.hp > 0:

                # wait a second. what if our enemy is dead..?
                if u.target.hp <= 0.:
                    # find a new random enemy.
                    target = find_nearest_enemy(u, units)
                    if target != -1:
                        u.target = target
                    else:
                        # no more enemies - stop!
                        running = False

                """Only do all this once we have a valid target..."""
                # euclidean distance
                u.dd = deriv_norm(u.pos, u.target.pos)
                u.dist = euclidean(u.pos, u.target.pos)
                # now, if the unit is not in range, then move towards it
                if u.dist > u.range:
                    # speed modifier added, others ignored from above.
                    u.pos += dudt(u.pos, u.target.pos, u.speed)
                else:
                    # perform damage to enemy health.
                    u.target.hp -= u.dmg

        # save
        steps.append(deepcopy(units))
        # add step
        t += 1

    return steps


"""############################## ANIMATE FUNCTIONS #########################################"""


def set_boundary(ax, pos):
    """Sets the boundary on an axis."""
    xmin, xmax = np.min(pos[:, 0, :]) - 1., np.max(pos[:, 0, :]) + 1.
    ymin, ymax = np.min(pos[:, 1, :]) - 1., np.max(pos[:, 1, :]) + 1.
    # set bounds
    ax.set_xlim(xmin, xmax)
    ax.set_ylim(ymin, ymax)
    # hide axes labels
    ax.get_xaxis().set_visible(False)
    ax.get_yaxis().set_visible(False)


def basic_animate1(results, n):
    """The first basic animation function."""
    # extract data points
    pos = extract_pos(results, n)
    plt.rcParams["animation.html"] = "html5"
    N_f = len(results)

    # define a figure
    fig = plt.figure(figsize=(8, 6))
    ax = fig.add_subplot(111)
    # define boundaries - we use a function
    set_boundary(ax, pos)

    # set the object to draw to
    u_points, = ax.plot([], [], 'x', color="black", alpha=.8, markersize=5.)

    fig.tight_layout()
    # close the plot
    plt.close()

    # define init
    def _init():
        u_points.set_data(pos[0, 0, :], pos[0, 1, :])
        return u_points

    # define animate
    def _animate(i):
        u_points.set_data(pos[i, 0, :], pos[i, 1, :])
        return u_points

    ani = animation.FuncAnimation(fig, _animate, init_func=_init, interval=100, frames=N_f, blit=False)
    plt.show()
    return ani


def basic_animate2(results, n):
    """The second basic animation function."""
    N_f = len(results)
    # extract data points
    pos = extract_pos(results, n)
    hp = extract_hp(results, n)
    T = extract_team(results, n)

    T_uniq = np.unique(T)
    plt.rcParams["animation.html"] = "html5"
    # define a figure
    fig = plt.figure(figsize=(8, 6))
    ax = fig.add_subplot(111)
    # define boundaries - we use a function
    set_boundary(ax, pos)

    # create a colormap to map our team to a color - use the colorwheel
    cm = dict(zip(T_uniq, colorwheel()))

    # set the objects to draw to - we set them into lists.
    points_alive = toflat(
        [ax.plot([], [], 'o', color=cm[a], alpha=.6, markersize=10.) for a in T_uniq])
    points_dead = toflat(
        [ax.plot([], [], 'x', color=cm[a], alpha=.4, markersize=5.) for a in T_uniq])

    fig.tight_layout()
    # close the plot
    plt.close()

    # define init
    def _init():
        for j, a in enumerate(T_uniq):
            T_Alive = np.argwhere((hp[:, 0] > 0) & (T[:, 0] == a)).flatten()
            T_Dead = np.argwhere((hp[:, 0] <= 0) & (T[:, 0] == a)).flatten()

            if len(T_Alive) > 0:
                points_alive[j].set_data(pos[0, 0, T_Alive], pos[0, 1, T_Alive])
            if len(T_Dead) > 0:
                points_dead[j].set_data(pos[0, 0, T_Dead], pos[0, 1, T_Dead])
        return ((*points_alive, *points_dead))

    # define animate
    def _animate(i):
        for j, a in enumerate(T_uniq):
            T_Alive = np.argwhere((hp[:, i] > 0) & (T[:, i] == a)).flatten()
            T_Dead = np.argwhere((hp[:, i] <= 0) & (T[:, i] == a)).flatten()

            if len(T_Alive) > 0:
                points_alive[j].set_data(pos[i, 0, T_Alive], pos[i, 1, T_Alive])
            if len(T_Dead) > 0:
                points_dead[j].set_data(pos[i, 0, T_Dead], pos[i, 1, T_Dead])

        return ((*points_alive, *points_dead))

    ani = animation.FuncAnimation(fig, _animate, init_func=_init, interval=100, frames=N_f, blit=False)
    plt.show()
    return ani


def basic_animate3(results, n):
    """The third basic animation function."""
    N_f = len(results)
    # extract data points
    pos = extract_pos(results, n)
    dpos = extract_dpos(results, n)
    hp = extract_hp(results, n)
    T = extract_team(results, n)

    T_uniq = np.unique(T)

    plt.rcParams["animation.html"] = "html5"

    # define a figure
    fig = plt.figure(figsize=(8, 6))
    ax = fig.add_subplot(111)
    # define boundaries - we use a function
    set_boundary(ax, pos)

    # create a colormap to map our team to a color - use the colorwheel
    cm = dict(zip(T_uniq, colorwheel()))

    # set the objects to draw to - we set them into lists.
    arrow_alive = []

    for j, a in enumerate(T_uniq):
        T_Alive0 = np.argwhere((hp[:, 0] > 0) & (T[:, 0] == a)).flatten()

        alive = ax.quiver(pos[0, 0, T_Alive0], pos[0, 1, T_Alive0],
                          dpos[0, 0, T_Alive0], dpos[0, 1, T_Alive0],
                          color=cm[a], alpha=.5, scale=30,
                          width=0.015, pivot="mid")
        arrow_alive.append(alive)

    points_dead = toflat(
        [ax.plot([], [], 'x', color=cm[a], alpha=.4, markersize=5.) for a in T_uniq])

    # figure things
    fig.tight_layout()
    # close the plot
    plt.close()

    # define init
    def _init():
        for _j, _a in enumerate(T_uniq):
            T_Alive = np.argwhere((hp[:, 0] > 0) & (T[:, 0] == _a)).flatten()
            T_Dead = np.argwhere((hp[:, 0] <= 0) & (T[:, 0] == _a)).flatten()

            if len(T_Alive) > 0:
                arrow_alive[_j].set_UVC(dpos[0, 0, T_Alive], dpos[0, 1, T_Alive])
            if len(T_Dead) > 0:
                points_dead[_j].set_data(pos[0, 0, T_Dead], pos[0, 1, T_Dead])
        return ((*arrow_alive, *points_dead))

    # define animate
    def _animate(i):
        for _j, _a in enumerate(T_uniq):
            # get IDs of alive and dead units.
            T_Alive = np.argwhere((hp[:, i] > 0) & (T[:, i] == _a)).flatten()
            T_Dead = np.argwhere((hp[:, i] <= 0) & (T[:, i] == _a)).flatten()

            if len(T_Alive) > 0:
                # we #'set the offsets' i.e arrow position.
                arrow_alive[_j].set_offsets(pos[i, :, T_Alive])
                # we set the directions of the arrows
                arrow_alive[_j].set_UVC(dpos[i, 0, T_Alive], dpos[i, 1, T_Alive])
            if len(T_Dead) > 0:
                points_dead[_j].set_data(pos[i, 0, T_Dead], pos[i, 1, T_Dead])

        return ((*arrow_alive, *points_dead))

    ani = animation.FuncAnimation(fig, _animate, init_func=_init, interval=100, frames=N_f, blit=False)
    plt.show()
    return ani


"""################################# EXTRACT FUNCTIONS ###################################"""


def extract_pos(results, n):
    """Extracting positions from results array."""
    return np.stack(([np.array([results[i][j].pos for i in range(len(results))]) for j in range(n)]), axis=2)


def extract_dpos(results, n):
    """Extracting change in positions from results array."""
    return np.stack(([np.array([results[i][j].dd for i in range(len(results))]) for j in range(n)]), axis=2)


def extract_hp(results, n):
    """Extracting HP from results array."""
    return np.vstack(([np.array([results[i][j].hp for i in range(len(results))]) for j in range(n)]))


def extract_team(results, n):
    """Extracting team number from results array."""
    return np.vstack(([np.array([results[i][j].team for i in range(len(results))]) for j in range(n)]))


"""########################## MATH FUNCTIONS ##############################################"""


def euclidean(u_i, u_j):
    """Determines the euclidean distance between ui and uj."""
    return np.sqrt(np.sum(np.square(u_i - u_j)))


def magnitude(p):
    """Determines the magnitude of a vector."""
    return np.sqrt(np.dot(p, p))


def deriv_norm(u_i, u_j):
    """Determines the normed-derivative between two vectors."""
    return (u_j - u_i) / euclidean(u_i, u_j)


def dudt(u_i, u_j, s_i):
    """Calculates the change between two vectors across time."""
    return s_i * 1. * 1. * deriv_norm(u_i, u_j)


"""####################### AI FUNCTION #######################################"""


def find_random_enemy(u, units):
    """Where u is the chosen unit, units is a list of all units."""
    # generate list of alive enemies
    alive_enemies = [i for i in units if (i.hp > 0) and (i.team != u.team)]
    # hang on, what if there are no alive enemies?
    if len(alive_enemies) > 0:
        return random.choice(alive_enemies)
    else:
        return -1


def find_nearest_enemy(u, units):
    """Finds the nearest enemy for a chosen unit."""
    # get the alive units, and calculate the euclidean distance from them.
    alive_enemies = [i for i in units if (i.hp > 0) and (i.team != u.team)]
    if len(alive_enemies) > 0:
        D = np.asarray([euclidean(u.pos, i.pos) for i in alive_enemies])
        return alive_enemies[np.argmin(D)]
    else:
        return -1


"""###################### IMPORT ################################"""


def check_columns(df, list_of_columns):
    for l in list_of_columns:
        if l not in df.columns:
            raise ValueError("column '{}' not found in dataframe.".format(l))


def import_db_file(fpath):
    """Imports our dataset of information, given a file path."""
    accepted_db_columns = (
        "Name", "Allegiance", "HP", "Damage", "Accuracy",
        "Miss", "Movement Speed", "Range"
    )

    if not os.path.isfile(fpath):
        raise IOError("fpath does not exist.")
    # attempt to read in
    df = pd.read_csv(fpath)
    # check columns exist
    check_columns(df, accepted_db_columns)
    df.set_index("Name", inplace=True)
    # assign an integer to allegiance
    df["allegiance_int"] = pd.factorize(df["Allegiance"])[0]
    return df


"""################### PLOTTING FUNCTIONS ################################"""


def plot_quiver(pos, dpos, legend_cap=5):
    """Basic plotting function for single quiver."""
    for i in range(pos.shape[-1]):
        if pos.shape[-1] > legend_cap:
            plt.scatter(pos[:, 0, i], pos[:, 1, i])
        else:
            plt.scatter(pos[:, 0, i], pos[:, 1, i], label="unit%d" % (i + 1))
        plt.quiver(pos[:, 0, i], pos[:, 1, i], dpos[:, 0, i], dpos[:, 1, i])

    plt.xlabel("x")
    plt.ylabel("y")
    if pos.shape[-1] <= legend_cap:
        plt.legend()
    plt.show()


"""####################### GROUP FUNCTIONS ########################################"""


def create_unit_group(df, u_type, n, pos_params):
    """
    Given a unit type, database, n sample and position params, create a unit group blob
    using gaussian.

    pos_params should be a tuple (2,) for mean and var.

    we cannot however assign enemies without a list to draw from, so this might have
    to be assigned after this function.
    """
    return [Unit1(df, u_type).set_position(pos_params) for _ in range(n)]


def assign_enemies(units):
    """Every unit gets assigned a random enemy."""
    # get team ids
    for u in units:
        u.target = find_random_enemy(u, units)


"""######################## MISC #######################################"""


def time_n(method, r=5, *args, **kws):
    """
    r for repeats
    """
    T = np.zeros(r + 1, )
    for i in range(r + 1):
        ts = time.time()
        _ = method(*args, **kws)
        T[i] = time.time() - ts
    # drop the first example as it is always slowest
    return T[1:]


def toflat(L):
    """
    Given nd-list, flatten to single dimension.
    """
    return list(it.chain.from_iterable(L))


def colorwheel():
    return ("red", "blue", "green", "orange", "purple", "brown", "black",
            "cyan", "yellow")



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Fri Feb 22 14:27:46 2019

@author: gparkes
"""
import os
import numpy as np
import itertools as it
import pandas as pd
import functools
import warnings
import time
from typing import Tuple, Iterable


def check_columns(df, list_of_columns):
    """Check that list of columns elements are not in df."""
    for l in list_of_columns:
        if l not in df.columns:
            raise ValueError("column '{}' not found in dataframe.".format(l))


def check_list_type(L, t):
    """Checks that every element in L is of type t."""
    for i, l in enumerate(L):
        if not isinstance(l, t):
            raise TypeError("type '{}' not found in list at index [{}]".format(t, i))
    return True


def colorwheel() -> Tuple:
    """All of the supported named-types in Matplotlib."""
    return ("red", "blue", "green", "orange", "purple", "brown", "black",
            "cyan", "yellow")


def check_in_list(L, sl: Iterable):
    """
    L is the reference list, sl is the 'sub list'
    """
    for i, l in enumerate(sl):
        if l not in L:
            raise ValueError("element '{}' not found in super-list {}".format(l, L))
    return True


def time_n(method, r=5, *args, **kws):
    """
    r for repeats
    """
    T = np.zeros(r + 1, )
    for i in range(r + 1):
        ts = time.time()
        _ = method(*args, **kws)
        T[i] = time.time() - ts
    # drop the first example as it is always slowest
    return T[1:]


def timed(method, *args, **kw):
    """Time a particular function."""
    ts = time.time()
    result = method(*args, **kw)
    te = time.time()
    diff = te - ts
    """
    Ranges : 0 to 80 (seconds)
    81 to 4800 (minutes)
    4801 to 100800 (hours)
    100800 + (days)
    """
    if diff < 80:
        out_str = "%r: %2.2f s"
    elif 80 <= diff < 4800:
        out_str = "%r: %2.2f m"
        diff /= 60.
    elif 4800 <= diff < 100800:
        out_str = "%r: %2.2f h"
        diff /= 3600.
    else:
        out_str = "%r: %2.2f d"
        diff /= 86400.

    print(out_str % (method.__name__, diff))
    return result


def timeit(method):
    """A decorator for timing methods."""
    def _timed(*args, **kw):
        ts = time.time()
        result = method(*args, **kw)
        te = time.time()
        diff = te - ts
        """
        Ranges : 0 to 80 (seconds)
        81 to 4800 (minutes)
        4801 to 100800 (hours)
        100800 + (days)
        """
        if diff < 80:
            out_str = "%r: %2.2f s"
        elif 80 <= diff < 4800:
            out_str = "%r: %2.2f m"
            diff /= 60.
        elif 4800 <= diff < 100800:
            out_str = "%r: %2.2f h"
            diff /= 3600.
        else:
            out_str = "%r: %2.2f d"
            diff /= 86400.

        print(out_str % (method.__name__, diff))
        return result

    return _timed


def is_twotuple(L, type1, type2):
    """
    Checks whether the object L is a 2-element tuple throughout.
    """
    if isinstance(L, (list, tuple)):
        for i, x in enumerate(L):
            if isinstance(x, tuple):
                if len(x) != 2:
                    raise ValueError("'L' element at index [{}] is not of size 2".format(i))
                if not isinstance(x[0], type1):
                    raise TypeError("x[0] element at index [{}] is not of type '{}'".format(i, type1))
                if not isinstance(x[1], type2):
                    raise TypeError("x[1] element at index [{}] is not of type '{}'".format(i, type2))
            else:
                raise TypeError("'x' is not of type 'tuple'")
        return True
    else:
        raise TypeError("'L' must be of type 'list, tuple'")


def is_ntuple(L, *types):
    """
    Checks whether the object L is an n-element tuple throughout.
    """
    if isinstance(L, (list, tuple)):
        if len(L) == len(types):
            for elem, t in zip(L, types):
                if not isinstance(elem, t):
                    raise TypeError("'L' element {} is not of type {}".format(elem, t))
        else:
            raise ValueError("L must be the same length as type list")
    else:
        raise TypeError("'L' must be of type [list, tuple]")


def get_segments(army_set: Tuple[str, int]):
    """
    Where army_set[i][0] is the unit type, army_set[i][1] is the N, (number of units)
    """
    s = [(0, army_set[0][1])]
    total = army_set[0][1]
    for i in range(1, len(army_set)):
        s.append((total, total + army_set[i][1]))
        total += army_set[i][1]
    return s


def check_groups_in_db(groups, db):
    for group_name, count in groups:
        if group_name not in db.index:
            raise ValueError("group '{}' not found in {}".format(group_name, db.index.tolist()))
    return True


def slice_loop(loopable, n):
    """Returns n elements from an infinite loop of loopable."""
    return list(it.islice(it.cycle(loopable), 0, n))


def max_norm(x):
    """Assumes x is a vector"""
    return x / np.max(x)


def sum_norm(x):
    return x / np.sum(x)


def io_table_columns():
    """Accepted column names for a unit file."""
    return (
        "Name", "Allegiance", "HP", "Armor", "Damage", "Accuracy",
        "Miss", "Movement Speed", "Range"
    )


def io_table_descriptions():
    """Descriptions corresponding to each accepted column name for a unit file."""
    return (
        "The name of the unit. Format string",
        "The team/allegiance of the unit. Format string, must be hashable",
        "HP: the health of the unit; an integer or float, no limit. Must be > 0",
        "Armor: the armor of the unit; an integer or float, no limit. Must be > 0",
        "Damage: the primary damage of the unit; integer or float, no limit.",
        "Accuracy: the accuracy of the unit; an integer/float in the range [0, 100]",
        "Miss: the chance of the unit to miss an attack; an integer/float in the range [0, 100]",
        "Movement Speed: the movement speed of the unit; float.",
        "Range: the range of the unit; integer or float, Must be > 0"
    )


def check_unit_file(df):
    """ Works inplace; no return """
    mappp = dict(zip(io_table_columns(), io_table_descriptions()))
    for m in io_table_columns():
        if m not in df.columns:
            raise IOError("column '{}' not found in file and must be present. Description:::'{}'".format(m, mappp[m]))


def preprocess_unit_file(df):
    """ Works inplace; no return """
    # assign index
    df.set_index("Name", inplace=True)
    # assign int allegiance
    df["allegiance_int"] = pd.factorize(df["Allegiance"])[0]


def import_and_check_unit_file(fpath):
    """
    Checks the quality of the unit-score datafile.

    Returns dataframe if successful.
    """
    # try to import
    if not os.path.isfile(fpath):
        raise IOError("fpath: {} does not exist".format(fpath))
    # attempt to read in
    df = pd.read_csv(fpath)
    # check
    check_unit_file(df)
    # preprocess
    preprocess_unit_file(df)

    return df



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Mon Jul  8 12:13:53 2019

@author: gparkes
"""
import numpy as np
from scipy import stats

__all__ = ["Distribution"]


class Distribution(object):
    """
    Here we provide a class that handles the creation, manipulation and deletion of
    Distributions which we can pass to manipulate the physical location of units
    in a given Battle.
    """

    def __init__(self, *args: str, **parameters):
        """
        Initialise the continuous distribution object with a name mapping to a
        scipy.stats distribution.

        Parameters
        --------
        args : list or Distribution
            Argument (in order) of the associated scipy.stats.<distribution>.
            Accepted parameters are the names of distribution ONLY.
            ['norm', 'uniform', 'gamma', 'beta', 'norm', 'laplace', 'exp', 'chi']
            Overrides if another distribution is in the parameters dict.
        parameters : dict
            Parameters of the associated scipy.stats.<distribution>, if wanted.
            Accepted keys: ['name', 'id', 'dist', 'distribution', 'mean', 'var',
                'loc', 'scale', 'a', 'b', 'df', 'sd', 'std']
        """
        self._dist = None

        if len(args) > 0:
            if len(args) == 1 and isinstance(args[0], Distribution):
                # copy
                self._dist = args[0]._dist
                self._x_param = args[0]._x_param
                self._y_param = args[0]._y_param
                self._option_i = args[0]._option_i
                self._dist_func = self.options_["funcs"][self._option_i]
                return
            # assuming we are using the standard arguments for scipy.stats
            self._get_distribution_name_from_args(*args)
        else:
            self._get_distribution_name_from_kwargs(**parameters)

        self._from_dict(**parameters)
        # get mapper ready
        self._dist_func = self.options_["funcs"][self._option_i]

    def _get_distribution_name_from_args(self, *args):
        if args[0] in self.options_["names"]:
            self.dist_ = args[0]
        elif args[0] in self.options_["name_maps"]:
            self.dist_ = self.options_["name_maps"][args[0]]
        else:
            raise ValueError("argument {} not found in options {}".format(args[0], self.options_["names"]))
        self._option_i = self.options_["names"].index(self.dist_)

    def _get_distribution_name_from_kwargs(self, **kws):
        # get the name
        # accepted keys
        if self.dist_ is not None:
            return
        else:
            dist_keys = ["name", "id", "dist", "distribution"]

            for k, arg in kws.items():
                if k in dist_keys:
                    if arg in self.options_["names"]:
                        self.dist_ = arg
                    elif arg in self.options_["name_maps"]:
                        self.dist_ = self.options_["name_maps"][arg]
                    else:
                        raise ValueError("argument '{}' not recognised in {}".format(arg, self.options_["names"]))

        if self.dist_ is not None:
            self._option_i = self.options_["names"].index(self.dist_)
        else:
            self.dist_ = "norm"
            self._option_i = 3

    def _from_dict(self, **kws):
        # arguments accepted
        dist_args = ["mean", "var", "loc", "scale", "sd", "std", "a", "b", "df"]
        dist_map = dict(zip(dist_args, ["loc", "scale", "loc", "scale", "scale", "scale", "a", "b", "df"]))
        d_xk, d_yk = {}, {}

        # iterate over them all
        for keyword, argument in kws.items():
            dim = "none"
            # get x_ y_ _x _y dims
            if keyword.startswith("x_"):
                keyword = keyword[2:]
                dim = "x"
            elif keyword.startswith("y_"):
                keyword = keyword[2:]
                dim = "y"
            elif keyword.endswith("_x"):
                keyword = keyword[:-2]
                dim = "x"
            elif keyword.endswith("_y"):
                keyword = keyword[:-2]
                dim = "y"

            if keyword in dist_args:
                # test argument
                if not isinstance(argument, (float, int, np.int, np.float, np.int64, np.float64)):
                    raise TypeError(
                        "argument '{}' must be of type [float, int] for distribution parameter".format(argument))

                k_m = dist_map[keyword]
                if dim == "x":
                    d_xk[k_m] = argument
                elif dim == "y":
                    d_yk[k_m] = argument
                else:
                    d_xk[k_m] = argument
                    d_yk[k_m] = argument

        # search to fill any missing gaps with default parameters
        must_params = self.options_["acc_params"][self._option_i]
        for p in must_params:
            if p not in d_xk:
                d_xk[p] = self.options_["def_params"][self._option_i][p]
            if p not in d_yk:
                d_yk[p] = self.options_["def_params"][self._option_i][p]

        # assign d_k to d.x_param_ - these throw errors if incorrect format
        self.x_param_ = d_xk
        self.y_param_ = d_yk

    @property
    def x_param_(self):
        """Returns all the x-parameters."""
        return self._x_param

    @x_param_.setter
    def x_param_(self, x):
        if not isinstance(x, dict):
            raise TypeError("'x' must be of type [dict]")
        if len(x) == 0:
            self._x_param = {}
            return
        for keyword, argument in x.items():
            if keyword not in self.options_["acc_params"][self._option_i]:
                raise ValueError("keyword '{}' does not belong in distribution '{}', which has {}"
                                 .format(keyword, self.dist_, self.options_["acc_params"][self._option_i]))
        self._x_param = x

    @property
    def y_param_(self):
        """Returns all of the y-parameters."""
        return self._y_param

    @y_param_.setter
    def y_param_(self, y):
        if not isinstance(y, dict):
            raise TypeError("'x' must be of type [dict]")
        if len(y) == 0:
            self._y_param = {}
            return
        for keyword, argument in y.items():
            if keyword not in self.options_["acc_params"][self._option_i]:
                raise ValueError("keyword '{}' does not belong in distribution '{}', which has {}"
                                 .format(keyword, self.dist_, self._options["acc_params"][self._option_i]))
        self._y_param = y

    def _get_dist_names(self):
        return self.options_["names"]

    @property
    def dist_(self):
        """The distribution type."""
        return self._dist

    @dist_.setter
    def dist_(self, dist):
        if not isinstance(dist, str):
            raise TypeError("dist must be of type [str]")
        if dist not in self.options_["names"]:
            raise ValueError("distribution '{}' not found in {}".format(dist, self.options_["names"]))
        self._dist = dist

    def _check_parameters(self, params):
        for p, v in params.items():
            if p not in self.options_["acc_params"][self._option_i]:
                raise ValueError(
                    "parameter '{}' not found in {}".format(p, self.options_["acc_params"][self._option_i]))
            if not isinstance(v, (float, np.float, np.float64, int, np.int, np.int64)):
                raise TypeError("parameter item must be of type [int, float]")

    @property
    def mean_(self):
        """The location parameters."""
        return np.asarray([self.x_param_["loc"], self.y_param_["loc"]])

    @property
    def options_(self):
        """A clunky dictionary with the total options."""
        return {
            "names": ["uniform", 'gamma', 'beta', 'norm', 'laplace', 'exp', 'chi'],
            "funcs": [stats.uniform, stats.gamma, stats.beta, stats.norm, stats.laplace,
                      stats.expon, stats.chi],
            # a dictionary to map alternative names used.
            "name_maps": {"gaussian": "norm", "normal": "norm", "expon": "exp", "unif": "uniform"},
            # default params
            "def_params": [
                {'loc': 0., 'scale': 1.}, {"a": 2., "loc": 0., "scale": 1.}, {"a": .5, "b": .5, "loc": 0., "scale": 1.},
                {"loc": 0., "scale": 1.}, {"loc": 0., "scale": 1.}, {"loc": 0., "scale": 1.},
                {"df": 1, "loc": 0., "scale": 1.}
            ],
            # accepted params
            "acc_params": [("loc", "scale"), ("a", "loc", "scale"),
                           ("a", "b", "loc", "scale"), ("loc", "scale"), ("loc", "scale"),
                           ("loc", "scale"), ("df", "loc", "scale")
                           ]
        }

    def setx(self, **x):
        """
        Set the x parameter if need be.
        """
        self._check_parameters(x)
        self.x_param_ = x
        return self

    def sety(self, **y):
        """
        Set the y parameter if need be.
        """
        self._check_parameters(y)
        self.y_param_ = y
        return self

    def sample(self, size):
        """
        Sample 'size' points from the x and y distribution, parameters.

        Parameters
        -------
        size : int
            N number of samples.

        Returns
        -------
        S : np.ndarray(size,2)
            S represents np.ndnarray matrix of positions.
        """
        S = np.zeros((size, 2), dtype=np.float_)
        S[:, 0] = self._dist_func(**self.x_param_).rvs(size=size)
        S[:, 1] = self._dist_func(**self.y_param_).rvs(size=size)
        return S

    def __repr__(self):
        return "{}: x={}, y={}".format(self.dist_, self.x_param_, self.y_param_)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Fri Feb 22 14:27:05 2019

@author: gparkes

A selection of algorithms for deciding which enemy to target next.

We use a rather ugly way of doing this - so we can use jit to speed up
the calculation using C rather than Python - all the numpy arrays are
passed to the functions.


    Parameters
    --------
    pos : np.ndarray (n, 2)
        The positions of all units
    hp : np.ndarray (n, )
        The HP of every unit.
    enemies : np.ndarray
        indices of enemy candidates
    allies : np.ndarray
        indices of ally candidates
    i : int
        Index of chosen unit

    Returns
    -------
    j : Index of new target
        -1 if not valid target chosen.

"""
import numpy as np
from numba import njit

from . import _jitcode


def get_function_names():
    return ["random", "nearest", "close_weak"]


def get_functions():
    return [random, nearest, close_weak]


def get_map_functions():
    return dict(zip(get_function_names(), get_functions()))


def get_global_function_names():
    return ["global_" + n for n in get_function_names()]


def get_global_functions():
    return [global_random, global_nearest, global_close_weak]


def get_global_map_functions():
    return dict(zip(get_global_function_names(), get_global_functions()))


__all__ = get_function_names() + get_global_function_names()


############## AI FUNCTIONS ##############################


@njit
def random(pos, hp, enemies, allies, i):
    """
    Given enemy candidates who are alive, draw one at random.
    """
    # draw a candidate
    if enemies.shape[0] > 0:
        return np.random.choice(enemies)
    else:
        return -1


@njit
def nearest(pos, hp, enemies, allies, i):
    """
    Given enemy candidates who are alive, determine which one is nearest.
    """
    if enemies.shape[0] > 0:
        # compute distances/magnitudes
        distances = _jitcode.euclidean_distance(pos[i] - pos[enemies])
        return enemies[np.argmin(distances)]
    else:
        return -1


@njit
def close_weak(pos, hp, enemies, allies, i, wtc_ratio=0.7):
    """
    Given enemy alive candidates, globally determine which one is the weakest
    and closest, using appropriate weighting for each option.

    Given enemy alive candidates, globally determine which one is the strongest
    of the enemies and target them.

    Parameters (extra)
    --------
    wtc_ratio : float [0..1]
        weak-to-close ratio to determine weighting of each part. Values closer
        to 1 prefer closer enemies, whereas values closer to 0 prefer weaker enemies
    """
    if enemies.shape[0] > 0:
        return enemies[
            np.argmin(
                (_jitcode.remove_mean(hp[enemies]) * (1. - wtc_ratio)) + \
                (_jitcode.remove_mean(_jitcode.euclidean_distance(pos[i] - pos[enemies])) * wtc_ratio)
            )
        ]
    else:
        return -1


######################### GLOBAL TARGET ASSIGNMENTS #######################################

"""
A selection of algorithms for deciding all enemies to target.

This is the same as above except there is no index parameter passed. Assumes
all units need a new target.


    Parameters
    --------
    pos : np.ndarray (n, 2)
        The positions of all units
    hp : np.ndarray (n, )
        The HP of every unit.
    team : np.ndarray (n, )
        Team number of every unit.
    group : np.ndarray (n, )
        The group number of every unit.
    group_i : int
        The group number selected

    Returns
    -------
    j : np.ndarray(n, )
        Index(es) of new target
"""


@njit
def global_random(pos, hp, team, group, group_i):
    # define
    selector = (group == group_i)
    t = np.unique(team[selector])[0]
    # get unit IDs that are not equal to this team for enemies.
    id_not, = np.where(team != t)
    # set the index for these guys
    j = np.random.choice(id_not, selector.sum())
    return j


@njit
def global_nearest(pos, hp, team, group, group_i):
    # define
    selector = (group == group_i)
    t = np.unique(team[selector])[0]
    # calculate distance matrix, with offset to ignore diagonal, with random noise
    D = _jitcode.distance_matrix(pos)
    D += np.eye(D.shape[0]) * np.max(D) + np.random.rand(D.shape[0], D.shape[0]) / 4.
    # get unit IDs that are not equal to this team for enemies.
    id_not, = np.where(team != t)
    id_is, = np.where(selector)
    # use distance matrix and ids to select sub groups to find argmin
    j = _jitcode.matrix_argmin(D[id_is, :][:, id_not])
    return j


@njit
def global_close_weak(pos, hp, team, group, group_i, wtc_ratio=0.7):
    # define
    selector = (group == group_i)
    t = np.unique(team[selector])[0]

    # calculate distance matrix, with offset to ignore diagonal, with random noise
    D = _jitcode.distance_matrix(pos)
    D += np.eye(D.shape[0]) * np.max(D) + np.random.rand(D.shape[0], D.shape[0]) / 4.

    # return the enemy that is closest and lowest HP
    hp_adj = _jitcode.remove_mean(hp) * (1. - wtc_ratio)
    dist_adj = _jitcode.remove_mean(D) * wtc_ratio

    # get unit IDs that are not equal to this team for enemies.
    id_not, = np.where(team != t)
    id_is, = np.where(selector)
    # use distance matrix and ids to select sub groups to find argmin
    j = _jitcode.matrix_argmin(dist_adj[id_is, :][:, id_not] + hp_adj[id_not])
    return j



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue Sep 24 18:07:16 2019

@author: gparkes

This file is concerned with movement functions associated with each army group.

There is the default 'move' and fancier options such as A*.
"""

from numba import njit

__all__ = ["to_enemy"]

"""
Parameters
    pos: (np.ndarray, (n, 2), float all) positions of every unit
    speed: (np.ndarray, (n,) float all) base speed of every unit
    dd: (np.ndarray, (n,) float all) directional derivative to target
    distances: (np.ndarray, (n,) float all) unit distance to target
    i : (int) the current unit's index
"""


@njit
def to_enemy(speed, dd, distances, Z, Z_xi, Z_yi, i):
    """
    Moves according to the euclidean distance towards the target.

    Returns adjustment, does not modify pos inplace.

    The update speed is calculated as:
        s_i * (dd_i / m_ij) * (1 - (Z_i / 2))
    """
    return (speed[i] * (dd[i] / distances[i])) * (1. - (Z[Z_xi[i], Z_yi[i]] / 2.))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue Oct 15 15:06:17 2019

@author: gparkes
"""

import numpy as np
from numba import njit


@njit
def _distance_matrix_calculation(X, D):
    """Determines the distance matrix (euclidean)."""
    for i in range(X.shape[0]):
        for j in range(i+1, X.shape[0]):
            D[i, j] = np.sqrt((X[i, 0] - X[j, 0])**2 + (X[i, 1] - X[j, 1])**2)


@njit
def _distance_vector_calculation(X, D):
    """Determines the distance as a vector."""
    for i in range(X.shape[0]):
        D[i] = np.sqrt((X[i, 0]*X[i, 0]) + (X[i, 1]*X[i, 1]))


@njit
def _matrix_min(X, A):
    """Matrix minimization function."""
    for i in range(X.shape[0]):
        A[i] = np.argmin(X[i, :])


@njit
def distance_matrix(X):
    """ Assumes X is 2-dimensional """
    D = np.zeros((X.shape[0], X.shape[0]), dtype=np.float64)
    # inplace op
    _distance_matrix_calculation(X, D)
    # flip D and add those values to it.
    D += D.transpose()
    return D


@njit
def matrix_argmin(X):
    """Calculates argmin along the axis=1 dimension """
    A = np.zeros(X.shape[0], dtype=np.int64)
    _matrix_min(X, A)
    return A


@njit
def boundary_check(bxmin, bxmax, bymin, bymax, pos):
    """performs boundary checks on our Mpos movement inplace"""
    pos[pos[:, 0] < bxmin, 0] = bxmin
    pos[pos[:, 0] > bxmax, 0] = bxmax
    pos[pos[:, 1] < bymin, 1] = bymin
    pos[pos[:, 1] > bymax, 1] = bymax


@njit
def euclidean_distance(M):
    """Determine the euclidean distance between all vectors in M."""
    D = np.zeros(M.shape[0])
    _distance_vector_calculation(M,D)
    return D


@njit
def direction_norm(dd):
    """Determine the directional norm between all vectors in dd."""
    distance = euclidean_distance(dd)
    return (dd.T/distance).T


@njit
def create_grid(xmin,xmax,ymin,ymax,res):
    """Creates a numpy mgrid given bounds."""
    return np.mgrid[xmin:xmax:res, ymin:ymax:res]


@njit
def minmax(X):
    """Calculate the minmax within vector X."""
    X_min = np.min(X)
    return (X - X_min) / (np.max(X) - X_min)


@njit
def remove_mean(x):
    """Removes the mean from vector x."""
    return x - np.mean(x)


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Mon Feb 25 13:38:51 2019

@author: gparkes
"""
import matplotlib.pyplot as plt
import numpy as np

__all__ = ["accuracy_contour", "hit_contour", "quiver_frame"]

"""################### PLOT FUNCTIONS HERE #######################################"""


def quiver_frame(xpd, frame_i=0):
    """
    Returns a single frame of quiver_fight.
    """
    # create plot
    fig = plt.figure(figsize=(8, 6))
    ax = fig.add_subplot(111)
    ax.set_xlim(xpd.x.min() - 1., xpd.x.max() + 1.)
    ax.set_ylim(xpd.y.min() - 1., xpd.y.max() + 1.)
    ax.set_xlabel("x")
    ax.set_ylabel("y")
    frame0 = xpd.query("(frame==@frame_i) & alive")
    ax.quiver(frame0.query("allegiance==0").x,
              frame0.query("allegiance==0").y,
              frame0.query("allegiance==0").dir_x,
              frame0.query("allegiance==0").dir_y,
              color="red", label="Republic", alpha=.5,
              scale=30, width=0.015, pivot="mid")
    ax.quiver(frame0.query("allegiance==1").x,
              frame0.query("allegiance==1").y,
              frame0.query("allegiance==1").dir_x,
              frame0.query("allegiance==1").dir_y,
              color="blue", label="CIS", alpha=.5, scale=30,
              width=0.015, pivot="mid")
    ax.legend(loc="best")
    fig.tight_layout()
    plt.show()
    return fig


def accuracy_contour(M, index, acc_penalty=15., boxlim=15):
    """
    Given a selected unit, with it's position, simulate accuracy/miss distances
    to draw accuracy contours.
    """
    xlim = np.asarray([M["pos"][index, 0] - boxlim, M["pos"][index, 0] + boxlim])
    ylim = np.asarray([M["pos"][index, 1] - boxlim, M["pos"][index, 1] + boxlim])

    xgrid = np.linspace(xlim[0], xlim[1], 100)
    ygrid = np.linspace(ylim[0], ylim[1], 100)
    X, Y = np.meshgrid(xgrid, ygrid)
    Z = np.sqrt(np.power(X - M["pos"][index, 0], 2) + np.power(Y - M["pos"][index, 1], 2))
    # normalize
    n_Z = (1. - (Z / acc_penalty)) * M["acc"][index]
    # get range also

    fig, ax = plt.subplots()
    ax.set_title("Normalized hit chance drop-off, ignoring enemy dodge.")
    ax.set_xlim(xlim)
    ax.set_ylim(ylim)

    ax.scatter(M["pos"][index, 0], M["pos"][index, 1], c="red")
    CS = ax.contour(X, Y, n_Z, levels=15)
    # add range circle
    ax.add_patch(plt.Circle((M["pos"][index, 0], M["pos"][index, 1]), M["range"][index], color="green", alpha=.3))
    ax.clabel(CS, inline=1, fontsize=8)
    return fig, ax


def hit_contour(M, i1, i2, acc_penalty=15., boxlim=10):
    """
    Given two units, simulate the contours for hitting each other.
    """

    def _contour_grid(M, j):
        xgrid = np.linspace(M["pos"][j, 0] - boxlim, M["pos"][j, 0] + boxlim, 100)
        ygrid = np.linspace(M["pos"][j, 1] - boxlim, M["pos"][j, 1] + boxlim, 100)
        X, Y = np.meshgrid(xgrid, ygrid)
        Z = np.sqrt(np.power(X - M["pos"][j, 0], 2) + np.power(Y - M["pos"][j, 1], 2))
        return X, Y, Z

    u1x, u1y, u1z = _contour_grid(M, i1)
    u2x, u2y, u2z = _contour_grid(M, i2)
    # normalize
    nZu1 = (1. - (u1z / acc_penalty)) * M["acc"][i1] * (1. - M["dodge"][i2])
    nZu2 = (1. - (u2z / acc_penalty)) * M["acc"][i2] * (1. - M["dodge"][i1])

    fig, ax = plt.subplots()
    # scatter
    ax.scatter(M["pos"][i1, 0], M["pos"][i1, 1], c="red", s=50)
    ax.scatter(M["pos"][i2, 0], M["pos"][i2, 1], c="green", marker="^", s=50)
    # add range circle
    ax.add_patch(plt.Circle((M["pos"][i1, 0], M["pos"][i1, 1]), M["range"][i1], color="green", alpha=.3))
    ax.add_patch(plt.Circle((M["pos"][i2, 0], M["pos"][i2, 1]), M["range"][i2], color="red", alpha=.3))
    # contours
    CS1 = ax.contour(u1x, u1y, nZu1, levels=10, cmap="winter")
    CS2 = ax.contour(u2x, u2y, nZu2, levels=10, cmap="summer")
    ax.clabel(CS1, inline=1, fontsize=6)
    ax.clabel(CS2, inline=1, fontsize=6)
    return fig, ax



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue Oct  1 15:40:34 2019

@author: gparkes

This class is for quantifying the usefulness of each unit.

This relies on the default columns being made available to it.
"""

import pandas as pd
import numpy as np

from . import utils


def rank_score(db, w=None):
    """
    Using the dataset, we can 'rank score' each unit according to it's relation
    to every other unit.

    We can do this uniformly over given features, or using weighting.

    We then sum the normalized ranks and take the average for a 'score' for each
    unit.

    Parameters
    -------
    db : pd.DataFrame
        The dataset in question.
    w : np.ndarray (4, ) [0..1]
        Weights for each rankable column. Currently six.
        [Damage, Range, Movement, HP]
    Returns
    -------
    p : pd.Series
        The scores for each unit.
    """
    if w is None:
        w = np.ones(4,)
    else:
        assert len(w) == 4, "len(w) must equal 6"

    rankable_columns = ["Damage","Range","Movement Speed", "HP"]
    # check they are there!
    utils.check_in_list(db.columns.tolist(), rankable_columns)

    Rank = pd.DataFrame(
        [utils.max_norm(db[r].rank()) * weight for weight, r in zip(w, rankable_columns)]
    )
    return Rank.mean().sort_values()



"""
Basic set up file
"""

from setuptools import setup, find_packages

with open("README.md","r") as fh:
    long_description = fh.read()

setup(
    name="battlesim",
    version="0.3.6",
    description="A python package for simulating battles and visualizing them in animation",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/gregparkes/battlesim",
    author="Gregory Parkes",
    author_email="g.m.parkes@soton.ac.uk",
    license="GPL-3.0",
    packages=find_packages(),
    zip_safe=False,
    python_requires='>=3.6',
    install_requires=[
        "numpy>=1.11.0",
        "scipy>=1.3",
        "pandas>=0.25.1",
        "matplotlib>=3.1.1",
        "numba>=0.45"
    ],
    classifiers=[
        "Natural Language :: English",
        "Programming Language :: Python :: 3.7",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Framework :: IPython",
        "Framework :: Jupyter",
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Science/Research",
        "Topic :: Scientific/Engineering",
    ],
)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed Oct  2 16:45:50 2019

@author: gparkes

This file determines different background 'terrains' to have for any given map.
This can be fixed size or infinite if it follows some mathematical function.
"""
import numpy as np
from matplotlib.pyplot import subplots
from scipy.stats import multivariate_normal
from typing import Optional, Tuple

from ._jitcode import minmax
from numba import njit
from . import utils


def get_tile_size(dim: Tuple[float, float, float, float],
                  res: float):
    """Returns the tile size of the resolution."""
    return int(np.abs(dim[0] - dim[1]) // res), int(np.abs(dim[2] - dim[3]) // res)


def _generate_random_gauss(pos: np.ndarray,
                           dim: Tuple[float, float, float, float]):
    def lm(x, sl, b):
        """Linear model function."""
        return x * sl + b

    # fetch dimensions
    xmin, xmax, ymin, ymax = dim
    Nx = (xmax - xmin)
    Ny = (ymax - ymin)
    # scaling factors
    # perfect scaling factor based on linear model y=mx + b
    x_n = np.arange(0, 300, 5)
    y_n = np.linspace(5, 3, 60)
    coef = np.polyfit(x_n, y_n, deg=1)
    # calculate scaling factor from formula
    sx = Nx / lm(Nx, coef[0], coef[1])
    sy = Ny / lm(Ny, coef[0], coef[1])

    print(sx, sy)
    # define mean
    m = np.random.rand(2) * np.array([Nx, Ny]) + np.array([xmin, ymin])
    # Diagonal elements for covariance matrix
    D = np.random.rand(2) + np.array([4., 1.5])
    # random choice to invert cov
    R = np.random.choice([1., -1.])
    # construct covariance
    C = np.array([[D[0] * sx, D[1] * R], [D[1] * R, D[0] * sy]])

    print(m, C)
    # use scipy.stats
    z = multivariate_normal(m, C).pdf(pos)
    return z


def _generate_noise_terrain(dim_x, dim_y):
    noise = np.random.rand(dim_x, dim_y)

    # %%
    @njit
    def _smooth_noise(x, y, noisewidth=100, noiseheight=100):
        # get fractional part
        fractX = x - int(x)
        fractY = y - int(y)
        # wrap
        x1 = (int(x) + noisewidth) % noisewidth
        y1 = (int(y) + noiseheight) % noiseheight
        # neighbor values
        x2 = (x1 + noisewidth - 1) % noisewidth
        y2 = (y1 + noiseheight - 1) % noiseheight
        # smooth the noise with bilinear interpolation
        val = 0.
        val += fractX * fractY * noise[y1, x1]
        val += (1 - fractX) * fractY * noise[y1, x2]
        val += fractX * (1 - fractY) * noise[y2, x1]
        val += (1 - fractX) * (1 - fractY) * noise[y2, x2]
        return val

    @njit
    def _turbulence(x, y, size):
        val = 0.
        init_size = size
        while size >= 1:
            val += smooth_noise(x / size, y / size, dim_x, dim_y) * size
            size /= 2.
        return 128 * val / init_size

    for x in range(dim_x):
        for y in range(dim_y):
            noise[x, y] = turbulence(x, y, 40)
    return noise


class Terrain(object):
    """
    The Terrain object is responsible for generating a terrain or map
    for the background of a given simulation.

    The Z-variable (depth) can be used in movement calculations of units, depending
    on which square they are walking on.

    In addition, the Z-variable could be used in calculations of the range/damage
    etc.

    The terrain object is responsible not only for the type of background but the
    way in which it is displayed.
    """

    def __init__(self,
                 dim: Tuple[float, float, float, float] = (0., 10., 0., 10.),
                 res: float = .1,
                 form: Optional[str] = "contour"):
        """
        Defines a Terrain object.

        Parameters
        -------
        dim : tuple (4,)
            x-min, x-max, y-min and y-max dimensions of the associated area.
        res : float
            The resolution of the map. Higher resolutions are slower but more accurate.
        form : str
            Determines what type of terrain to generate. Choose from ['grid', 'contour']
                -grid applies a function z = f(x, y) or random and discretizes this into square boxes.
                -contour applies a function z = f(x, y) or random and plots filled-contours of the background.
                -None defines a flat grid with no height penalties.
        """
        self.form_ = form
        self.bounds_ = dim
        self.res_ = res
        self._Z = None

    """ '################################## ATTRIBUTES ########################################### """

    @property
    def form_(self):
        """Determines the appearance of the terrain. Either grid or contour."""
        return self._form

    @form_.setter
    def form_(self, f: Optional[str]):
        if f not in [None, "grid", "contour"]:
            raise AttributeError("'form' must be [None, grid, contour]")
        self._form = f

    @property
    def bounds_(self):
        """Determines the dimensions by which to calculate Terrain."""
        return self._bounds

    @bounds_.setter
    def bounds_(self, dim: Tuple[float, float, float, float]):
        if not isinstance(dim, (list, tuple)):
            raise TypeError("'dim' must be of type [list, tuple]")
        if len(dim) != 4:
            raise AttributeError("'dim' must contain 4 elements")
        if dim[1] <= dim[0]:
            raise AttributeError("xmax cannot be <= xmin")
        if dim[3] <= dim[2]:
            raise AttributeError("ymax cannot be <= ymin")
        utils.is_ntuple(dim, *([(int, float, np.float, np.int, np.float32, np.int32)] * 4))
        self._bounds = dim

    @property
    def res_(self):
        """The resolution of the Terrain."""
        return self._res

    @res_.setter
    def res_(self, r: float):
        if not isinstance(r, (float, np.float)):
            raise TypeError("'res' must be of type [float]")
        if r < 1e-8:
            raise ValueError("'res' cannot be less than 0")
        self._res = r

    @property
    def Z_(self):
        """The array defining height."""
        return self._Z

    """############################## HIDDEN FUNCTIONS ################################################"""

    def _m_size(self):
        return (int((self.bounds_[1] - self.bounds_[0]) / self.res_),
                int((self.bounds_[3] - self.bounds_[2]) / self.res_))

    def _generate_random_terrain(self, n_gauss: int = 3):
        # define meshgrid
        X, Y = self.get_grid()

        # create empty positions
        pos = np.empty(X.shape + (2,))
        pos[:, :, 0] = X
        pos[:, :, 1] = Y

        Z_cont = np.stack(([_generate_random_gauss(pos, self.bounds_) for _ in range(n_gauss)]), axis=2)

        # add probabilities
        Z = Z_cont.sum(axis=2)
        # scale between 0 and 1 and return
        return minmax(Z)

    def __repr__(self):
        return "Terrain(init={}, type='{}', dim={}, resolution={:0.3f})".format(
            self.Z_ is not None, self.form_, self.bounds_, self.res_)

    """##################################### FUNCTIONS ##################################################"""

    def get_grid(self):
        """Returns the grid as an mgrid."""
        return np.mgrid[self.bounds_[0]:self.bounds_[1]:self.res_,
               self.bounds_[2]:self.bounds_[3]:self.res_]

    def get_flat_grid(self):
        """Produces a flat terrain grid."""
        X, Y = self.get_grid()
        # they are repeats, return single.
        return X[:, 0], Y[0, :]

    def generate(self, f=None, n_random: int = 50):
        """
        Generates the terrain using a function.

        Parameters
        ----------
        f : function or None
            If None, uses 'random', else uses a map_function if ['grid', 'contour']
        n_random : int, optional
            The number of random gaussians to generate.

        Returns
        -------
        Z : np.ndarray
            The contour grid for the plane.
        """
        if self.form_ is None:
            self._Z = np.zeros(self._m_size())
        if f is None:
            self._Z = self._generate_random_terrain(n_random)
        elif callable(f):
            self._Z = minmax(f(*self.get_grid()))
        else:
            raise TypeError("'f' must be a function or None")
        return self

    def plot(self, ax=None, **kwargs):
        """Plots the terrain as a contour to visualize."""
        # given an axes, plot the terrain using the parameters given.
        if self.Z_ is None:
            raise ValueError("Terrain not instantiated, call generate()")
        if ax is None:
            fig, ax = subplots(figsize=(8, 6))

        if self.form_ == "grid":
            ax.imshow(self.Z_, aspect="auto", cmap="binary", extent=self.bounds_, **kwargs)
        elif self.form_ == "contour" or self.form_ == "random":
            X, Y = self.get_grid()
            ax.contourf(X, Y, self.Z_, cmap="binary", extent=self.bounds_, **kwargs)
        return
